/***********************************************************************\
*
* Level 9 interpreter
* Version 4.0
* Copyright (c) 1996 Glen Summers
* Copyright (c) 2002,2003,2005 Glen Summers and David Kinder
*
* Level9 32 bit Windows version by Glen Summers and David Kinder.
* Windows CE version by Nikita Borodikhin
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*
\***********************************************************************/

#include <app.h>
#include <main.h>
#include <window.h>
#include <commctrl.h>
#include <filedlg.h>
#include <dialog.h>
#include <regstuff.h>

#include <fname.h>
#include <string.h>
#include <wstring.h>
#include <screenstring.h>

#include <level9.h>
#include <screen.h>
#include <inputoutput.h>

#include <sipapi.h>

#include <aygshell.h>

// define application name, main window title
#define AppName "Level9"
#define MainWinTitle "Level9 Interpreter"
#define DefaultFontFace "Tahoma"
 
#include "resource/resource.h"

#define BLINK_INTERVAL_MSEC 300

#define MAX_FONTFACES 20

#define FONTSIZEMIN   8
#define FONTSIZEMAX   20
#define FONTSIZESTEP  1

#define CLEARTYPE_QUALITY	5
#define MENUBAR_HEIGHT		26

// Aygshell imports
static BOOL has_aygshell = 0;

typedef BOOL WINAPI tShSipInfo (UINT uiAction, UINT uiParam, PVOID pvParam, UINT fWinIni);
static tShSipInfo *impSHSipInfo;

typedef BOOL tShCreateMenuBar (SHMENUBARINFO *pmb);
static tShCreateMenuBar *impSHCreateMenuBar;


String LogDir = "\\Temp";
char Ini[] = "Software\\Level9";

const wchar_t Filters[]=L"Level 9 Game Files (*.dat)\0*.dat\0Spectrum Snapshots (*.sna)\0*.sna\0All Files (*.*)\0*.*\0\0";
const wchar_t GameFilters[]=L"Saved game file (*.sav)\0*.sav\0All Files (*.*)\0*.*\0\0";
int FiltIndex;
int GameFiltIndex;

String LogFile;
FName LastFile;
FName LastGameFile;

int PageWidth, TextHeight, PageStart, PageFullHeight;
int FontHeight, LineSpacing = 1;
int Margin;
int Cursorx,Cursory;
int cbar_height;
int ShowSIP;
int ShowPicture = 1;
int ScalePicture = 0;
int ShowBlinker = 0;
int WriteLog = 1;
int UseCleartype = 0;
int UseFontBold = 0;
int UseFontItalic = 0;

/* Amiga */
RGBQUAD AmigaColours [8] = {
	{ 0x00,0x00,0x00,0 }, { 0x00,0x00,0xFF,0 },
	{ 0x30,0xE8,0x30,0 }, { 0x00,0xFF,0xFF,0 },
	{ 0xFF,0x00,0x00,0 }, { 0x00,0x68,0xA0,0 },
	{ 0xFF,0xFF,0x00,0 }, { 0xFF,0xFF,0xFF,0 }};
/* Shalaev ZX-Spectrum Emulator */
RGBQUAD SpectrumColours [8] = {
	{ 0x00,0x00,0x00,0 }, { 0x00,0x00,0xFC,0 },
	{ 0x00,0xFC,0x00,0 }, { 0x00,0xFC,0xFC,0 },
	{ 0xFC,0x00,0x00,0 }, { 0xF0,0x00,0xFC,0 },
	{ 0xFC,0xFC,0x00,0 }, { 0xFC,0xFC,0xFC,0 }};
int ColourScheme = 0; // Amiga

COLORREF Palette [32];
int PaletteIndex [8];

HWND hWndMain=0;
HRGN hClip=0;
int GfxMode = 0;
int GfxWidth = 0;
int GfxHeight = 0;
int GfxPosX = 0;
int GfxPosY = 0;
int GfxPicWidth=0, GfxPicHeight=0;

typedef int GRAPHTYPE;
GRAPHTYPE **Graph = NULL;
int GraphHeight = 0;

HBITMAP hGfx=0;
//HBITMAP hGfxDraw=0;
HPALETTE hPalette, hPal;
HDC hGfxDC=0;
//HDC hGfxDrawDC=0;

LOGFONT lf;
HFONT Font=0;
COLORREF FontColour;

void DrawPicture(void);
void ShowSip (BOOL Show);
void ShowHideSIP ();
void SetFontProps ();

void DrawOneLine (int x, int y, int dx, int dy, const String &line)
{
	if (x < 0) return;

	HDC dc=GetDC(hWndMain);
	HFONT OldFont=(HFONT)SelectObject(dc,Font);
	COLORREF OldCol=SetTextColor(dc,FontColour);
	COLORREF OldBk=SetBkColor(dc,GetSysColor(COLOR_WINDOW));
	WS wline = line;
	if(hClip) SelectClipRgn(dc,hClip);
	ExtTextOut(dc, Margin + x, PageStart + y * LineSpacing + (FontHeight < 0 ? -FontHeight : 0),
				0, NULL, wline,wcslen (wline),NULL);
	SelectObject(dc,OldFont);
	SetTextColor(dc,OldCol);
	SetBkColor(dc,OldBk);
	ReleaseDC(hWndMain,dc);
}

void UnDrawRect (int x, int y, int dx, int dy)
{
	RECT rc;
	rc.left=Margin + x;
	rc.right=Margin + x + dx;
	rc.top=PageStart + y * FontHeight;
	rc.bottom=PageStart + (y + dy) * FontHeight;
	HDC dc=GetDC(hWndMain);
	if(hClip) SelectClipRgn(dc,hClip);
	FillRect(dc,&rc,(HBRUSH) GetStockObject (WHITE_BRUSH));
	ReleaseDC(hWndMain,dc);
}

int LineLength(char*str,int n)
{
	SIZE Size;
	HDC dc=GetDC(hWndMain);
	HFONT OldFont=(HFONT)SelectObject(dc,Font);
	String sstr = str;
	{
		WS wstr=sstr;
		GetTextExtentPoint32(dc,wstr,wcslen (wstr),&Size);
	}
	SelectObject(dc,OldFont);
	ReleaseDC(hWndMain,dc);
	return Size.cx;
}

void MakeCaret()
{
	if (GetFocus()==hWndMain && Caret)
	{
		CreateCaret(hWndMain, NULL, 2, FontHeight);
		SetCaretPos(Cursorx,Cursory);
		ShowCaret();
	}
}

void KillCaret()
{
	if (Caret) DestroyCaret();
}

void ShowCaret ()
{
	ShowCaret (hWndMain);
}

void HideCaret ()
{
	HideCaret (hWndMain);
}

void SetCaret(int x,int y)
{
	Cursorx = Margin + x;
	Cursory = PageStart + y * LineSpacing;
	SetCaretPos(Cursorx, Cursory);
}

void Redraw()
{
	DrawStrings ();
	DrawPicture ();
}

void PreScrollWindow ()
{
	RECT rect;

	rect.top = PageStart + LineSpacing;
	rect.bottom = PageFullHeight;
	rect.left = Margin;
	rect.right = PageWidth; // in case if too long line
	
	ScrollWindowEx(hWndMain, 0, -LineSpacing,
		           &rect, NULL, NULL, NULL, SW_ERASE);

	rect.top = PageFullHeight - LineSpacing;
	rect.bottom = PageFullHeight;
	rect.left = 0;
	rect.right = PageWidth;

	HDC dc=GetDC(hWndMain);
	FillRect(dc,&rect,(HBRUSH) GetStockObject (WHITE_BRUSH));
	ReleaseDC(hWndMain,dc);
}

void PostScrollWindow () {}

void PostAddScreenString () {}

void Paginate()
{
	RecreateScreenLines ();

	if (Caret)
	{
		HideCaret ();
		SetLastCaret();
		ShowCaret ();
	}
}

void Resize()
{
	if (!has_aygshell)
		GfxPosY = cbar_height;

	if (GfxMode && ShowPicture && GfxPicHeight && GfxPicWidth)
	{
		if (!ScalePicture)
		{
			GfxHeight = GfxPicHeight;
			GfxWidth  = GfxPicWidth;

		}
		else
		{
			GfxPosX = 0;
			GfxWidth = PageWidth;
			GfxHeight = (GfxPicHeight * PageWidth) / GfxPicWidth;
			//GfxHeight = GfxPicHeight;
		}

		if (PageFullHeight - GfxPosY - GfxHeight < 2 * LineSpacing)
		{
			int oldheight = GfxHeight;
			GfxHeight = PageFullHeight - GfxPosY - 2 * LineSpacing;

			if (GfxHeight <= 0)
			{
				GfxHeight = 0;
				GfxWidth = 0;
			}
			else
				GfxWidth = (GfxWidth * GfxHeight) / oldheight;
		}

		GfxPosX = (PageWidth - GfxWidth) / 2; // centering
	}
	else
	{
		GfxPosX = 0;
		GfxHeight = 0;
		GfxWidth  = 0;
	}

	PageStart = GfxPosY + GfxHeight;

	if (FontHeight < 0)
		PageStart += abs (FontHeight);
	TextHeight = PageFullHeight - PageStart;
	ScreenString::MaxWidth = PageWidth - 2 * Margin;
	ScreenString::MaxLines = TextHeight / LineSpacing;

	if (hClip)
		DeleteObject(hClip);

	hClip=CreateRectRgn(0,PageStart,PageWidth,PageFullHeight);

	Paginate();
	InvalidateRect(hWndMain,NULL,TRUE);
	HideCaret ();
	MakeCaret ();
}

void Wait(int millis)
{
	static int blink_colour = -1;
	static int blink_mtime  = 0;

	if (hGfxDC)
	{
		int gfx_status = TRUE;
		int gfx_count = 0;
		int gfx_limit = 20*millis;

		while (gfx_status && (gfx_count < gfx_limit))
		{
			gfx_status = RunGraphics();
			if (gfx_status)
				gfx_count++;
		}

		if (gfx_count > 0)
			DrawPicture();
	}

	if (blink_mtime > BLINK_INTERVAL_MSEC)
	{
		blink_mtime = 0;
		blink_colour = (blink_colour && ShowBlinker) ? 0 : 0xff;

		static POINT lines [7] = {{0,0}, {0,0}, {1,0}, {1,0}, {2,0}, {2,0}, {0,0}};
		HDC dc;
		HPEN Pen, OldPen;

		lines [0].y = PageFullHeight - 1;
		lines [1].y = PageFullHeight - 5;
		lines [2].y = PageFullHeight - 1;
		lines [3].y = PageFullHeight - 5;
		lines [4].y = PageFullHeight - 1;
		lines [5].y = PageFullHeight - 5;
		lines [6].y = PageFullHeight - 1;

		dc = GetDC (hWndMain);

		Pen = CreatePen (PS_SOLID, 1, RGB (blink_colour, blink_colour, blink_colour));
		OldPen = (HPEN) SelectObject (dc, Pen);

		Polyline (dc, lines, 7);
		DeleteObject (Pen);
		SelectObject (dc, OldPen);

		ReleaseDC (hWndMain, dc);
	}

	blink_mtime += millis;

	Sleep(millis);
}

char os_readchar (int millis)
{
  int c;
  
  c = os_getchar (millis);

  if ((c < 0) || (c > 255)) return 0;

  return (char (c));
}

int os_getchar (int millis)
{
	int count = 0;
	int result = 0;

	SimpleNode<int> *C;

	while (millis >= 0 && ((C = InputChars.GetFirst ()) == NULL) &&
		   !(result = App::PeekMessage ()) &&
		   ((C = InputChars.GetFirst ()) == NULL))
	{
		if (millis > BLINK_INTERVAL_MSEC)
			Wait (BLINK_INTERVAL_MSEC);
		else
			Wait (millis);

		millis -= BLINK_INTERVAL_MSEC;
	}

	if (!result && (C == NULL))
		return KEY_TIMEOUT;

	if ((C = InputChars.GetFirst ()) == NULL)
		return 0;

	int c = C->Item;
	delete C;

	switch (c)
	{
		case 256+VK_DOWN:
			return KEY_DOWN_CODE;
		case 256+VK_UP:
			return KEY_UP_CODE;
		case 256+VK_LEFT:
			return KEY_LEFT_CODE;
		case 256+VK_RIGHT:
			return KEY_RIGHT_CODE;
		case 256+VK_END:
			return KEY_END_CODE;
		case 256+VK_HOME:
			return KEY_HOME_CODE;
		case 8:
			return KEY_BACKSPACE_CODE;
		case 256+VK_DELETE:
			return KEY_DELETE_CODE;
		case '\r':
		case '\n':
			return '\r';
		case -2: // ReplaceInput
			return -2;
	}
	if ((c >= 32) && (c <= 127))
		return c;
  
	return -1;
}

L9BOOL os_stoplist(void)
{
	return (os_readchar(0) != 0);
}

void os_set_filenumber(char *NewName,int Size,int n)
{
	FName fn(NewName);
	String S;
	fn.GetBaseName(S);
	while ((S.Last () >= '0') && (S.Last () <= '9')) S.Remove(S.Len()-1,1);
	fn.NewBaseName(S << n);
	strcpy(NewName,fn);
}

BOOL os_get_game_file(char* Name,int Size)
{
	String SName (Size);
	SName = Name;
	int sip = ShowSIP;
	ShowSIP = 0;
	ShowHideSIP ();

	int result = OpenFileDlg(hWndMain, Filters, &FiltIndex,
				             SName, Size, "Open Game File");

	strcpy (Name, (char *) SName);

	ShowSIP = sip;
	ShowHideSIP ();

	return result;
}

BOOL os_save_file(BYTE *Ptr,int Bytes)
{
	CancelInput();
	BOOL result;
	int sip = ShowSIP;

	ShowSIP = 0;
	ShowHideSIP ();

	result = FALSE;

	if (SaveFileDlg (hWndMain, GameFilters, &GameFiltIndex,
		             LastGameFile, LastGameFile.Size(), "Save File"))
	{
		LastGameFile.Update();
   		if (!LastGameFile.GetExt()) LastGameFile.NewExt("sav");

		HANDLE fhand = CreateFile (WS (LastGameFile), GENERIC_WRITE, 0,
			NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 0);
		if (fhand)
		{
			DWORD written;

			WriteFile (fhand, Ptr, Bytes, &written, NULL);

			if ((int) written != Bytes)
				MessageBox (hWndMain, WS (String () << "Error while writing to file " << LastGameFile),
							L"Write Error",MB_OK | MB_ICONEXCLAMATION);
			else
				result = TRUE;

			CloseHandle (fhand);
		}
		else
			MessageBox (hWndMain, WS (String () << "Couldn't open file " << LastGameFile << " for write"),
						L"Write Error",MB_OK | MB_ICONEXCLAMATION);

	}
	
	ShowSIP = sip;
	ShowHideSIP ();

	return result;
}

BOOL os_load_file(BYTE *Ptr,int *Bytes,int Max)
{
	int sip = ShowSIP;
	BOOL result;

	ShowSIP = 0;
	ShowHideSIP ();

	CancelInput();

	result = FALSE;

	if (OpenFileDlg(hWndMain, GameFilters, &GameFiltIndex,
				    LastGameFile, LastGameFile.Size (), "LoadFile"))
	{
		LastGameFile.Update();

		HANDLE fhand = CreateFile (WS (LastGameFile), GENERIC_READ, 0,
			NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 0);

		if (fhand)
		{
			DWORD read;
			DWORD size = GetFileSize (fhand, NULL);
			*Bytes = size;

			if (size == 0xFFFFFFFF) 
				MessageBox (hWndMain, WS (String () << "Error occured while opening file " << LastGameFile << " for read"),
							L"Read Error",MB_OK | MB_ICONEXCLAMATION);
			else if (*Bytes > Max)
				MessageBox (hWndMain, WS (String () << LastGameFile << " is not a valid saved game file"),
							L"Invalid file",MB_OK | MB_ICONEXCLAMATION);
			else if (!ReadFile (fhand, Ptr, *Bytes, &read, NULL))
				MessageBox (hWndMain, WS (String () << "Error occured while opening file " << LastGameFile << " for read"),
							L"Read Error",MB_OK | MB_ICONEXCLAMATION);
			else
			{
				result = TRUE;
			}

			CloseHandle (fhand);
		}
		else
			MessageBox (hWndMain, WS (String () << "Couldn't open file " << LastGameFile << " for read"),
						L"Write Error",MB_OK | MB_ICONEXCLAMATION);
	}

	ShowSIP = sip;
	ShowHideSIP ();

	return result;
}

void SetIndexPalette(RGBQUAD* pal)
{
	int i;

	switch (GfxMode)
	{
		case 1:
			RGBQUAD *Colours;

			switch (ColourScheme)
			{
				case 1:		Colours = SpectrumColours; break;
				default:	Colours = AmigaColours;
			}

			for (i = 0; i < 8; i++)
			{
				pal[i].rgbRed = Colours [i].rgbRed;
				pal[i].rgbBlue = Colours [i].rgbBlue;
				pal[i].rgbGreen = Colours [i].rgbGreen;
			}

			break;
		case 2:
			for (i = 0; i < 32; i++)
			{
				pal[i].rgbRed	= GetRValue (Palette [i]);
				pal[i].rgbBlue	= GetBValue (Palette [i]);
				pal[i].rgbGreen = GetGValue (Palette [i]);
			}

			break;
		default:
			;
	}
}

void DrawPicture(void)
{
	if (hGfxDC)
	{
		HDC dc = GetDC(hWndMain);
		StretchBlt(dc, GfxPosX, GfxPosY, GfxWidth, GfxHeight,
			hGfxDC, 0, 0, GfxPicWidth, GfxPicHeight, SRCCOPY);
		ReleaseDC (hWndMain, dc);
	}
}

FName GfxDir;
BitmapType GfxBmapType = NO_BITMAPS;
BYTE* GfxBits = NULL;
int LastBitmap = -1;

void CreateGFXDC ()
{
	typedef struct tagBITMAPINFO32
	{
		BITMAPINFOHEADER bmiHeader; 
		RGBQUAD bmiColors[32];
	}
	BITMAPINFO32;

	HDC dc = GetDC(hWndMain);
	hGfxDC = CreateCompatibleDC(dc);
	ReleaseDC(hWndMain,dc);

	BITMAPINFO32 info;
	memset (&info, 0, sizeof(BITMAPINFO32));
	info.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	info.bmiHeader.biPlanes = 1;
	info.bmiHeader.biBitCount = 8;
	info.bmiHeader.biCompression = BI_RGB;
	info.bmiHeader.biWidth = GfxPicWidth;
	info.bmiHeader.biHeight = -GfxPicHeight;
	info.bmiHeader.biClrUsed = 32;
	SetIndexPalette(info.bmiColors);

	hGfx = CreateDIBSection(hGfxDC,(BITMAPINFO*)&info,DIB_RGB_COLORS,(void **) &GfxBits,NULL,0);
	SelectObject(hGfxDC,hGfx);
}

void os_graphics(int mode)
{
	GfxBmapType = NO_BITMAPS;
	LastBitmap = -1;

	switch (mode)
	{
	case 0:
		GfxMode = 0;
		break;
	case 1:
		GfxMode = 1;
		break;
	case 2:
		LastFile.GetDir(GfxDir);
		GfxDir.AddChar('\\');
		GfxBmapType = DetectBitmaps(GfxDir);
		GfxMode = (GfxBmapType != NO_BITMAPS) ? 2 : 0;
		break;
	}

	if (hGfx)
		DeleteObject(hGfx);
	hGfx = 0;
	GfxBits = NULL;
	if (hGfxDC)
		DeleteDC(hGfxDC);
	hGfxDC = 0;
	if (Graph && GraphHeight)
	{
		for (int i = 0; i < GraphHeight; i++)
			delete [] Graph [i];
		delete [] Graph;
		GraphHeight = 0;
	}
	Graph = NULL;

	if (GfxMode)
	{
		RGBQUAD *Colours;
		int i;

		if (GfxMode == 2)
		{
			GfxPicWidth = 0;
			GfxPicHeight = 0;
		}
		else
		{
			GetPictureSize(&GfxPicWidth,&GfxPicHeight);
		}

		CreateGFXDC ();

		if (GfxMode == 1)
		{
			switch (ColourScheme)
			{
				case 1:		Colours = SpectrumColours; break;
				default:	Colours = AmigaColours; break;
			}

			for (i = 0; i < 8; i++)
				Palette [i] = RGB (Colours [i].rgbRed, Colours [i].rgbGreen, Colours [i].rgbBlue);

			GraphHeight = GfxPicHeight;
			Graph = new GRAPHTYPE* [GraphHeight];

			for (i = 0; i < GfxPicHeight; i++)
				Graph [i] = new GRAPHTYPE [GfxPicWidth];

			os_cleargraphics ();
		}
	}

	Resize();
}

void os_cleargraphics(void)
{
	if (hGfxDC && GfxMode == 1)
	{
		int i, j;
		RECT rc;

		for (i = 0; i < 8; i++)
			PaletteIndex [i] = i;

		rc.left = 0;
		rc.right = GfxPicWidth;
		rc.top = 0;
		rc.bottom = GfxPicHeight;

		HBRUSH br = CreateSolidBrush(Palette [PaletteIndex [0]]);
		HGDIOBJ old = SelectObject(hGfxDC,br);
		FillRect(hGfxDC,&rc,br);
		SelectObject(hGfxDC,old);
		DeleteObject(br);

		for (i = 0; i < GfxPicHeight; i++)
			for (j = 0; j < GfxPicWidth; j++)
				Graph [i][j] = 0;
	}
}

COLORREF colour_plot;
int colour_set;
int colour_replace;

void SetPoint (int x, int y)
{
	Graph [y][x] = colour_set;
	SetPixel (hGfxDC, x, y, colour_plot);
}

void PointIfColor (int x, int y)
{
	if ((x >= 0) && (x < GfxPicWidth) && (y >= 0) && (y < GfxPicHeight) &&
		(Graph [y][x] == colour_replace))
		SetPoint (x, y);
}

void LineIfColor(int x1, int y1, int x2, int y2)
{
	register int i, dx, dy, dxabs, dyabs, x, y, px, py;
	register signed char sdx, sdy;

	dx = x2 - x1;      /* the horizontal distance of the line */
	dy = y2 - y1;      /* the vertical distance of the line */
	dxabs = abs (dx);
	dyabs = abs (dy);
	sdx = dx ? ((dx > 0) ? 1 : -1) : 0;
	sdy = dy ? ((dy > 0) ? 1 : -1) : 0;
	x = dyabs / 2;
	y = dxabs / 2;
	px = x1;
	py = y1;

	PointIfColor (px, py);

	if (dxabs >= dyabs) /* the line is more horizontal than vertical */
	{
		for (i = 0; i < dxabs; i++)
		{
			px += sdx;
			PointIfColor (px, py);
			y += dyabs;
			if (y >= dxabs)
			{
				y -= dxabs;
				py += sdy;
			}
		}
	}
	else /* the line is more vertical than horizontal */
	{
		for (i = 0; i < dyabs; i++)
		{
			py += sdy;
			PointIfColor (px, py);
			x += dxabs;
			if (x >= dyabs)
			{
				x -= dyabs;
				px += sdx;
			}
		}
	}
	PointIfColor (px, py);
}

void os_drawline(int x1, int y1, int x2, int y2, int colour1, int colour2)
{
	if (hGfxDC)
	{
		colour_set  = colour1;
		colour_plot = Palette [PaletteIndex [colour_set]];
		colour_replace = colour2;
		LineIfColor (x1, y1, x2, y2);
	}
}

void FloodFillRegion (int x, int y);

void os_fill(int x, int y, int colour1, int colour2)
{
	if (colour1 == colour2) return;

	if (!((x >= 0) && (x < GfxPicWidth) && (y >= 0) && (y < GfxPicHeight)))
		return;

	if (hGfxDC)
	{
		colour_replace = colour2;

		if (Graph [y][x] == colour_replace)
		{
			colour_set = colour1;
			colour_plot = Palette [PaletteIndex [colour_set]];
			HPEN CurPen = CreatePen (PS_SOLID, 1, Palette [PaletteIndex [colour_set]]);
			HGDIOBJ OldPen = SelectObject (hGfxDC, CurPen);
			FloodFillRegion (x, y);
			SelectObject (hGfxDC, OldPen);
			DeleteObject (CurPen);
		}
	}
}

inline void HorizLine (int x1, int x2, int y)
{
	static POINT points [2];

	points [0].x = x1;
	points [0].y = y;
	points [1].x = x2 + 1;
	points [1].y = y;

	Polyline (hGfxDC, points, 2);
}

void FloodFillRegion (int x, int y)
{
	register int lastBorder;
	register int old;
	register int leftLimit, rightLimit;
	register int i;

	old = Graph [y][x];
	if ((old == colour_set) || (old != colour_replace))
		return;

	/* Seek left */
	leftLimit = (-1);
	for (i = x; i >= 0; i--)
	{
		if (Graph [y] [i] != colour_replace)
			break;
		Graph [y][i] = colour_set;
		leftLimit = i;
	}

	if (leftLimit == (-1))
		return;

	HorizLine (leftLimit, x, y);

	/* Seek right */
	rightLimit = x;
	for (i = (x + 1); i < GfxPicWidth; i++)
	{
		if (Graph [y] [i] != colour_replace)
			break;
		Graph [y] [i] = colour_set;
		rightLimit = i;
	}

	HorizLine (x, rightLimit, y);

	/* Look at lines above and below and start paints */
	/* Above */
	if (y > 0)
	{
		lastBorder = 1;
		for (i = leftLimit; i <= rightLimit; i++)
		{
			register GRAPHTYPE c;
			c = Graph [y-1][i];
			if (lastBorder)
			{
				if (c == old)
				{
					FloodFillRegion (i, y - 1);
					lastBorder = 0;
				}
			}
			else
			if (c != old)
				lastBorder = 1;
		}
	}
	/* Below */
	if (y < GfxPicHeight - 1)
	{
		lastBorder = 1;
		for (i = leftLimit; i <= rightLimit; i++)
		{
			register GRAPHTYPE c;
			c = Graph [y+1][i];
			if (lastBorder)
			{
				if (c == old)
				{
					FloodFillRegion (i, y + 1);
					lastBorder = 0;
				}
			}
			else
			if (c != old)
				lastBorder = 1;
		}
	}
}

void os_setcolour(int colour, int index)
{
	if (hGfxDC)
	{
		PaletteIndex [colour] = index;

		int startx;
		HPEN CurPen= CreatePen (PS_SOLID, 1, Palette [index]);
		HGDIOBJ OldPen = SelectObject (hGfxDC, CurPen);

		for (int i = 0; i < GfxPicHeight; i++)
		{
			startx = -1;

			for (int j = 0; j < GfxPicWidth; j++)
			{
				if (Graph [i][j] == colour)
				{
					if (startx == -1)
						startx = j;
				}
				else
				{
					if (startx != -1)
						HorizLine (startx, j - 1, i);
					startx = -1;
				}
			}
			if (startx != -1)
				HorizLine (startx, j - 1, i);
		}

		SelectObject (hGfxDC, OldPen);
		DeleteObject (CurPen);
	}
}

void os_show_bitmap(int pic, int x, int y)
{
	int i;

	if (LastBitmap == pic)
		return;

	if ((GfxMode == 2))
	{
		Bitmap* bitmap = DecodeBitmap(GfxDir,GfxBmapType,pic,x,y);
		if (bitmap)
		{
			if ((GfxPicWidth  != bitmap->width) ||
				(GfxPicHeight != bitmap->height))
			{
				GfxPicWidth  = bitmap->width;
				GfxPicHeight = bitmap->height;

				Resize ();
			}

			for (i = 0; i < bitmap->npalette; i++)
			{
				Palette [i] = RGB (bitmap->palette [i].red,
									bitmap->palette [i].green,
									bitmap->palette [i].blue
						);
			}

			CreateGFXDC ();

			if (GfxBits)
			{
				int wcepadding = 0;
				int wceoffset = 0;
				int srcoffset = 0;
				
				if (GfxPicWidth % sizeof (DWORD))
					wcepadding = (sizeof (DWORD) - GfxPicWidth % sizeof (DWORD));

				for (i = 0; i < GfxPicHeight; i++)
				{
					memcpy (GfxBits + wceoffset,
							bitmap->bitmap + srcoffset,
							GfxPicWidth);
					
					srcoffset += GfxPicWidth;
					wceoffset += GfxPicWidth + wcepadding;
				}

				DrawPicture();
			}

			LastBitmap = pic;

			if ((pic == 0) && (ShowPicture))
				::Sleep(1000);
		}
	}
}

void ConstructLogFile ()
{
	String temp;

	LastFile.GetBaseName (temp);

	LogFile = LogDir;

	LogFile << "\\";
	LogFile << temp << ".log";
}

void os_write_log_string (String &str)
{
	if (WriteLog)
	{
		String temp;
		char ch;

		FILE *file = fopen (LogFile, "a");

		for (int i = 0; i < str.Length; i++)
			switch (ch = str [i])
			{
				case '\r': temp << "\n"; break;
				case '%':  temp << "%%"; break;
				case '\\': temp << "\\\\"; break;
				default:  temp << ch;
			}
		temp << "\n";

		fprintf (file, temp);

		fclose (file);
	}
}

// About Dialog ******************************************

class AboutDialog : public Dialog
{
	public:
		AboutDialog(Object *Parent) : Dialog(Parent,IDD_ABOUT,"AboutDialog") {}
		virtual BOOL CanClose () {return TRUE;}
};

// MainWindow *****************************************

Array<String> FontFamilies;

class MainWindow : public Window
{
public:
	MainWindow(Object *Parent,char *Title);
	~MainWindow();
	HWND cbar;
	static HMENU OptionsMenu;
	HMENU FileMenu, HelpMenu;
	HMENU FontMenu, FontSizeMenu;
	HICON hIcon, hSmallIcon;

	BOOL Playing;

	void Destroy();
	BOOL SetupWindow();
	void OpenFile(FName name);

	void CmAbout();
	void CmExit();
	void CmOpen();
	void CmSelectFont (int id);
	void CmSelectFontSize (int size);
	void CmSIP ();
	void CmScale ();
	void CmPicture ();
	void CmAbbrev ();
	void CmRestore() { ReplaceInput ("#restore"); }
	void CmSave() { ReplaceInput ("save"); }
	void CmDictionary() { ReplaceInput ("#dictionary"); }
	void CmAbbreviations() { ReplaceInput ("#abbreviations"); }
	void CmSchemeAmiga ();
	void CmSchemeSpectrum ();
	void CmWriteLog ();
	void CmChooseLogDir ();
	void CmShowBlinker ();
	void CmCleartype ();
	void CmFontBold ();
	void CmFontItalic ();

	void SetFont ();
	void DelFonts ();
	void UpdateFont () {DelFonts (); SetFontProps (); SetFont (); Resize ();};

	virtual BOOL WndProc(TMSG &Msg);
	virtual LRESULT DefProc(HWND hWnd,UINT Msg,WPARAM wParam,LPARAM lParam);

// message response functions

	BOOL LButtonDown(TMSG &);
	BOOL RButtonDown(TMSG &);
	BOOL LButtonUp(TMSG &);
	BOOL WMMouseMove(TMSG &);
	BOOL WMSize(TMSG &);

	BOOL WMKeyDown(TMSG &);
	BOOL WMChar(TMSG &);
	BOOL WMSetFocus(TMSG&);
	BOOL WMKillFocus(TMSG&);
	BOOL WMSettingChange(TMSG&);

// window paint request
	void Paint(HDC, BOOL, RECT&);

// enable message response
	HASH_EV_ENABLE(MainWindow)
} ;

// define response functions
EV_START(MainWindow)
// command messages
	EV_COMMAND(CM_ABOUT, CmAbout)
	EV_COMMAND(CM_EXIT, CmExit)
	EV_COMMAND(CM_OPEN, CmOpen)
	EV_COMMAND(CM_SIP, CmSIP)
	EV_COMMAND(CM_SCALE, CmScale)
	EV_COMMAND(CM_PICTURE, CmPicture)
	EV_COMMAND(CM_ABBREV, CmAbbrev)
	EV_COMMAND(CM_FILELOAD, CmRestore)
	EV_COMMAND(CM_FILESAVE, CmSave)
	EV_COMMAND(CM_DICTIONARY, CmDictionary)
	EV_COMMAND(CM_ABBREVIATIONS, CmAbbreviations)
	EV_COMMAND(CM_SCHEME_AMIGA, CmSchemeAmiga)
	EV_COMMAND(CM_SCHEME_SPECTRUM, CmSchemeSpectrum)
	EV_COMMAND(CM_WRITELOG, CmWriteLog)
	EV_COMMAND(CM_LOGDIR, CmChooseLogDir)
	EV_COMMAND(CM_SHOWBLINKER, CmShowBlinker)
	EV_COMMAND(CM_CLEARTYPE, CmCleartype)
	EV_COMMAND(CM_FONT_BOLD, CmFontBold)
	EV_COMMAND(CM_FONT_ITALIC, CmFontItalic)

// windows messages
	EV_MESSAGE(WM_LBUTTONDOWN, LButtonDown)
	EV_MESSAGE(WM_KEYDOWN, WMKeyDown)
	EV_MESSAGE(WM_CHAR, WMChar)
	EV_MESSAGE(WM_SIZE, WMSize)
	EV_MESSAGE(WM_SETFOCUS,WMSetFocus)
	EV_MESSAGE(WM_KILLFOCUS,WMKillFocus)
// SIP message
	EV_MESSAGE(WM_SETTINGCHANGE, WMSettingChange)
EV_END

HMENU MainWindow::OptionsMenu = 0;

// main window constructor
MainWindow::MainWindow(Object *Parent,char *Title) : Window(Parent,Title,"MainWindow")
{
	// set window style
	Style=0/*WS_OVERLAPPEDWINDOW*/;
	Icon = IDI_ICON;
}

static int CALLBACK EnumFontFams (const LOGFONT *lpelf,
						   const TEXTMETRIC FAR *lpntm,
							unsigned long FontType,
							LPARAM lParam )
{
	if (lpelf->lfCharSet == ANSI_CHARSET)
	{
		char *face = StringFromWS (lpelf->lfFaceName);
		String *sface = new String (face);

		FontFamilies.Add (sface);
	}

	return 1;
}

// this is called to setup the window
BOOL MainWindow::SetupWindow()
{
	//SetWindowLong(hWnd,GWL_EXSTYLE,WS_EX_OVERLAPPEDWINDOW);
	//SetWindowLong(hWnd,GWL_STYLE,WS_BORDER);

	//AssignMenu(IDM_MENU);
	//SetIcon(IDI_ICON);
	HMENU Menu;

	hWndMain=hWnd;
        
	if (has_aygshell)
	{
		SHMENUBARINFO mbi;

		memset(&mbi, 0, sizeof(SHMENUBARINFO));
		mbi.cbSize		= sizeof(SHMENUBARINFO);
		mbi.hwndParent	= hWnd;
#ifdef SHBMBF_HMENU
		mbi.dwFlags		= 0x0010; // For unknown reason MSDN includes
								  // SHBMBF_HMENU flag for SHMENUBARINFO.dwFlags
								  // but PPC SDK does not
#endif
		mbi.nToolBarId	= IDM_MENU;
		mbi.hInstRes	= App::hInstance;
		mbi.nBmpId		= 0;
		mbi.cBmpImages	= 0;	

		if (!(*impSHCreateMenuBar)(&mbi))
			MessageBox (hWnd, L"Error creating menu", L"Sorry", MB_OK);
	  
		cbar = mbi.hwndMB;
		Menu = (HMENU) SendMessage (cbar, SHCMBM_GETMENU, (WPARAM)0, (LPARAM)0);
	}
	else
	{
		cbar = CommandBar_Create (App::hInstance, hWnd, ID_CBAR);
		CommandBar_InsertMenubar (cbar, App::hInstance, IDM_MENU, 0);
		Menu = CommandBar_GetMenu (cbar, 0);
		CommandBar_AddAdornments (cbar, 0, 0);
	}

	FileMenu = GetSubMenu    (Menu, 0);
	OptionsMenu = GetSubMenu (Menu, 1);
	HelpMenu = GetSubMenu    (Menu, 2);

	// Hide caption
	Style&=~( (WS_CAPTION & ~WS_BORDER) | WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_SYSMENU);
	SetWindowLong(hWnd,GWL_STYLE,Style);

	hIcon = (HICON) LoadImage (App::hInstance, MAKEINTRESOURCE (IDI_ICON), IMAGE_ICON, 32, 32, 0);
	hSmallIcon = (HICON) LoadImage (App::hInstance, MAKEINTRESOURCE (IDI_ICON), IMAGE_ICON, 16, 16, 0);

	ShowHideSIP ();
	CheckMenuItem (OptionsMenu, CM_SCALE,
		MF_BYCOMMAND | (ScalePicture ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_PICTURE,
		MF_BYCOMMAND | (ShowPicture ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_ABBREV,
		MF_BYCOMMAND | (ExpandAbbreviations ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_SCHEME_AMIGA,
		MF_BYCOMMAND | (ColourScheme == 0 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_SCHEME_SPECTRUM,
		MF_BYCOMMAND | (ColourScheme == 1 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_WRITELOG,
		MF_BYCOMMAND | (WriteLog == 1 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_SHOWBLINKER,
		MF_BYCOMMAND | (ShowBlinker == 1 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_CLEARTYPE,
		MF_BYCOMMAND | (UseCleartype == 1 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_FONT_BOLD,
		MF_BYCOMMAND | (UseFontBold == 1 ? MF_CHECKED : MF_UNCHECKED));
	CheckMenuItem (OptionsMenu, CM_FONT_ITALIC,
		MF_BYCOMMAND | (UseFontItalic == 1 ? MF_CHECKED : MF_UNCHECKED));

	SetFont();
	Playing=FALSE;

	HDC dc = GetDC (hWnd);
	EnumFontFamilies (dc, NULL, EnumFontFams, 0);
	ReleaseDC (hWnd, dc);


	int i;
	int max = FontFamilies.GetSize ();
	if (max > CM_FONTFAM)
		max = CM_FONTFAM;

	FontMenu = CreatePopupMenu ();

	for (i = 0; i < max; i++)
	{
		if (!wcscmp (WS (*FontFamilies [i]), lf.lfFaceName))
			AppendMenu (FontMenu, MF_CHECKED | MF_STRING, CM_FONTFAM + i, WS (*FontFamilies [i]));
		else
			AppendMenu (FontMenu, MF_STRING, CM_FONTFAM + i, WS (*FontFamilies [i]));
	}

	AppendMenu (OptionsMenu, MF_POPUP, (UINT) FontMenu, L"Font Face");

	FontSizeMenu = CreatePopupMenu ();
	wchar_t buf [10];

	for (i = FONTSIZEMIN; i < FONTSIZEMAX; i += FONTSIZESTEP)
	{
		swprintf (buf, L"%i", i);

		if (abs (lf.lfHeight) == i)
			AppendMenu (FontSizeMenu, MF_CHECKED | MF_STRING, CM_FONTSIZE + i, buf);
		else
			AppendMenu (FontSizeMenu, MF_STRING, CM_FONTSIZE + i, buf);
	}
	AppendMenu (OptionsMenu, MF_POPUP, (UINT) FontSizeMenu, L"Font Size");

	return TRUE;
}

// this is called when the window is destroyed
void MainWindow::Destroy()
{
	DestroyIcon (hSmallIcon);
	DestroyIcon (hIcon);
	SipShowIM (SIPF_OFF);
	DelFonts();
	StopGame();
	Playing=FALSE;
	FreeMemory();
	CancelInput();
}

MainWindow::~MainWindow()
{
}

LRESULT MainWindow::DefProc(HWND hWnd,UINT Msg,WPARAM wParam,LPARAM lParam)
{
	if (Msg == WM_GETICON)
		return long (wParam ? hIcon : hSmallIcon);

	return Window::DefProc (hWnd, Msg, wParam, lParam);
}

BOOL MainWindow::WndProc(TMSG &Msg)
{
	int id = LOWORD (Msg.wParam);

	switch (Msg.Msg)
	{
		case WM_COMMAND:
			if ((id >= CM_FONTFAM) &&
				(id < CM_FONTFAM + MAX_FONTFACES))
			{
				CmSelectFont (id - CM_FONTFAM);
				return TRUE;
			}
			if ((id >= CM_FONTSIZE + FONTSIZEMIN) &&
				(id < CM_FONTSIZE + FONTSIZEMAX))
			{
				CmSelectFontSize (id - CM_FONTSIZE);
				return TRUE;
			}
		default:
			;
	}
	return Window::WndProc (Msg);
}

void MainWindow::CmSelectFont(int id)
{
	int i = FontFamilies.GetSize ();

	while (i--)
	{
		if (i == id)
			CheckMenuItem (FontMenu, i, MF_BYPOSITION | MF_CHECKED);
		else
			CheckMenuItem (FontMenu, i, MF_BYPOSITION | MF_UNCHECKED);
	}

	wcscpy (lf.lfFaceName, WS (*FontFamilies [id]));
	UpdateFont ();
}

void MainWindow::CmSelectFontSize (int size)
{
	int i;

	for (i = FONTSIZEMIN; i < FONTSIZEMAX; i += FONTSIZESTEP)
	{
		if (i == size)
			CheckMenuItem (FontSizeMenu, CM_FONTSIZE + i, MF_BYCOMMAND | MF_CHECKED);
		else
			CheckMenuItem (FontSizeMenu, CM_FONTSIZE + i, MF_BYCOMMAND | MF_UNCHECKED);
	}

	lf.lfHeight = -size;
	UpdateFont ();
}

void MainWindow::SetFont()
{
	char string [] = " ";
	SIZE size;

	Font=CreateFontIndirect(&lf);
	HDC dc=GetDC(hWnd);
	HFONT OldFont=(HFONT)SelectObject(dc,Font);
	TEXTMETRIC tm;
	GetTextMetrics(dc,&tm);
	Margin=tm.tmAveCharWidth;

	string [0] = ' ';
	GetTextExtentPoint32 (dc, WS (string), 1, &size);

	FontHeight = tm.tmHeight;
	LineSpacing = abs ((int) 1.1 * FontHeight);
	ScreenString::SpaceWidth = size.cx;

	SelectObject(dc,OldFont);
	ReleaseDC(hWnd,dc);
}

void MainWindow::DelFonts()
{
	DeleteObject(Font);
	Font = 0;
}


void ShowHideSIP ()
{
	static int oldShowSIP = -1;

	if (ShowSIP == oldShowSIP)
		return;

	oldShowSIP = ShowSIP;

	if (ShowSIP)
	{
		CheckMenuItem (MainWindow::OptionsMenu, CM_SIP, MF_BYCOMMAND | MF_CHECKED);

		ShowSip (TRUE);
	}
	else
	{
		CheckMenuItem (MainWindow::OptionsMenu, CM_SIP, MF_BYCOMMAND | MF_UNCHECKED);

		ShowSip (FALSE);
	}
}

void MainWindow::CmSIP ()
{
	ShowSIP = !ShowSIP;

	ShowHideSIP ();
}

void MainWindow::CmScale ()
{
	ScalePicture = !ScalePicture;
	CheckMenuItem (OptionsMenu, CM_SCALE,
		MF_BYCOMMAND | (ScalePicture ? MF_CHECKED : MF_UNCHECKED));

	Resize ();
}

void MainWindow::CmPicture ()
{
	ShowPicture = !ShowPicture;
	CheckMenuItem (OptionsMenu, CM_PICTURE,
		MF_BYCOMMAND | (ShowPicture ? MF_CHECKED : MF_UNCHECKED));

	Resize ();
}

void MainWindow::CmSchemeAmiga ()
{
	ColourScheme = 0;
	CheckMenuItem (OptionsMenu, CM_SCHEME_AMIGA, MF_BYCOMMAND | MF_CHECKED);
	CheckMenuItem (OptionsMenu, CM_SCHEME_SPECTRUM, MF_BYCOMMAND | MF_UNCHECKED);

	os_graphics (GfxMode);
}

void MainWindow::CmSchemeSpectrum ()
{
	ColourScheme = 1;
	CheckMenuItem (OptionsMenu, CM_SCHEME_AMIGA, MF_BYCOMMAND | MF_UNCHECKED);
	CheckMenuItem (OptionsMenu, CM_SCHEME_SPECTRUM, MF_BYCOMMAND | MF_CHECKED);

	os_graphics (GfxMode);
}

void MainWindow::CmAbbrev ()
{
	ExpandAbbreviations = !ExpandAbbreviations;
	CheckMenuItem (OptionsMenu, CM_ABBREV,
		MF_BYCOMMAND | (ExpandAbbreviations ? MF_CHECKED : MF_UNCHECKED));
}

void MainWindow::CmWriteLog ()
{
	WriteLog = !WriteLog;
	CheckMenuItem (OptionsMenu, CM_WRITELOG,
		MF_BYCOMMAND | (WriteLog ? MF_CHECKED : MF_UNCHECKED));
}

void MainWindow::CmChooseLogDir ()
{
	String temp (101);
	
	temp = LogDir;
	if (SaveDirDlg (hWnd, temp, 100, "Log file directory", "Logs will be placed here"))
	{
		LogDir = temp;
		ConstructLogFile ();
	}
}

void MainWindow::CmShowBlinker ()
{
	ShowBlinker = !ShowBlinker;
	CheckMenuItem (OptionsMenu, CM_SHOWBLINKER,
		MF_BYCOMMAND | (ShowBlinker ? MF_CHECKED : MF_UNCHECKED));
}

void MainWindow::CmCleartype ()
{
	UseCleartype = !UseCleartype;
	CheckMenuItem (OptionsMenu, CM_CLEARTYPE,
		MF_BYCOMMAND | (UseCleartype ? MF_CHECKED : MF_UNCHECKED));

	UpdateFont ();
}

void MainWindow::CmFontBold ()
{
	UseFontBold = !UseFontBold;
	CheckMenuItem (OptionsMenu, CM_FONT_BOLD,
		MF_BYCOMMAND | (UseFontBold ? MF_CHECKED : MF_UNCHECKED));

	UpdateFont ();
}

void MainWindow::CmFontItalic ()
{
	UseFontItalic = !UseFontItalic;
	CheckMenuItem (OptionsMenu, CM_FONT_ITALIC,
		MF_BYCOMMAND | (UseFontItalic ? MF_CHECKED : MF_UNCHECKED));

	UpdateFont ();
}

void MainWindow::OpenFile(FName name)
{
	// in input routine?, cause fall through
	CancelInput();
	// clear buffers ect..., enen if fails as invalidaes game memory
	InvalidateRect(hWnd,NULL,TRUE);

	// look for a picture datafile
	FName picname(name);
	picname.NewExt("pic");
	if (!picname.Exist()) picname.NewExt("hrc");
	if (!picname.Exist()) picname.NewExt("cga");
	if (!picname.Exist()) picname.NewName("picture.dat");

	if (!LoadGame(name,picname)) MessageBox(hWnd,L"Unable to load game file",L"Load Error",MB_OK | MB_ICONEXCLAMATION);

	else
	{
		LastFile=name;
		ConstructLogFile ();
		os_graphics (0);
		Redraw ();

		if (Playing) return;
		Playing=TRUE;
		while (Playing && RunGame()) App::PeekLoop();
		Playing=FALSE;
	}
}

void MainWindow::CmOpen()
{
	FName fn=LastFile;

	if (os_get_game_file(fn,fn.Size()))
		OpenFile(fn);
}

void MainWindow::CmExit()
{
	DestroyWindow();
}

void MainWindow::Paint(HDC, BOOL, RECT&)
{
	Redraw();
	UpdateWindow (cbar);
}

void MainWindow::CmAbout()
{
	AboutDialog(this).Execute();
}

BOOL MainWindow::WMKeyDown(TMSG &Msg)
{
	switch ((int) Msg.wParam)     // the virtual key code
	{
		case VK_F2: CmOpen(); break;
		case VK_F3: CmRestore(); break;
		case VK_F4: CmSave(); break;
		case VK_F5: CmDictionary(); break;
			
		case VK_LEFT:
		case VK_RIGHT:
		case VK_UP:
		case VK_DOWN:
		case VK_DELETE:
		case VK_END:
		case VK_HOME:
			InputChars.AddTail(256+Msg.wParam);
			break;
		}
	return TRUE; // message handled
}

BOOL MainWindow::WMChar(TMSG &Msg)
{
	InputChars.AddTail(Msg.wParam);
	return TRUE;
}

BOOL MainWindow::LButtonDown(TMSG &Msg)
{
	int xPos = LOWORD(Msg.lParam);  /* horizontal position of cursor */
	int yPos = HIWORD(Msg.lParam);  /* vertical position of cursor   */

	AddWordAtPos (xPos - Margin,
					(yPos - PageStart) / LineSpacing);

	return TRUE;
}

BOOL MainWindow::WMSize(TMSG &Msg)
{
	PageWidth = LOWORD(Msg.lParam);
	PageFullHeight = HIWORD(Msg.lParam);

	if (has_aygshell)
	{
		SendMessage(cbar, TB_AUTOSIZE, 0L, 0L);
	}
	else
	{
		// CommandBar_AlignAdornments(cbar);
	}

	if (has_aygshell)
	{
		cbar_height = 0;

		SIPINFO si;
		memset (&si, 0, sizeof (si));
		si.cbSize = sizeof (si);

		if ((*impSHSipInfo) (SPI_GETSIPINFO, 0, &si, 0) && !(si.fdwFlags & SIPF_ON))
		{
			RECT cbar_rect;
			GetWindowRect (cbar, &cbar_rect);
			cbar_height = cbar_rect.bottom - cbar_rect.top;
		}
	}
	else
	{
		cbar_height = CommandBar_Height (cbar);
	}

	if (has_aygshell)
	{
		PageFullHeight -= cbar_height;
	}

	UpdateFont ();
	return TRUE;
}

BOOL MainWindow::WMSetFocus(TMSG&)
{
	ShowSip (ShowSIP);

	MakeCaret();
	return TRUE;
}

BOOL MainWindow::WMKillFocus(TMSG&)
{
	KillCaret();

	ShowSip (FALSE);

	return TRUE;
}

BOOL MainWindow::WMSettingChange(TMSG& Msg)
{
	switch (Msg.wParam) 
	{
		case SPI_SETSIPINFO:
			SIPINFO si;
			memset (&si, 0, sizeof (si));
			si.cbSize = sizeof (si);

			if (SipGetInfo (&si)) 
			{
				SetWindowPos (hWnd, 0,
					si.rcVisibleDesktop.left,
					si.rcVisibleDesktop.top,
					si.rcVisibleDesktop.right - si.rcVisibleDesktop.left,
					si.rcVisibleDesktop.bottom - si.rcVisibleDesktop.top,
					SWP_NOOWNERZORDER | SWP_NOZORDER);
				ShowSIP = si.fdwFlags & SIPF_ON;
				ShowHideSIP ();
			}
			
			break;
		default:
			;
	}

	return 0;
}

// App *****************************************

class MyApp : public App
{
public:
	MyApp(char *Name);
	~MyApp();
	int w, h;
	
private:
	void InitMainWindow();
	void SetDefs();
	void ReadIni();
	void WriteIni();
	void FirstIn();
};

void MyApp::SetDefs()
{
	WS wface = DefaultFontFace;
	// Set application default settings here
	FiltIndex=0;
	LastGameFile="";
	GameFiltIndex=0;

	memset(&lf,0,sizeof(lf));
	lf.lfHeight=-12;
	lf.lfWeight=FW_NORMAL;
	lf.lfCharSet=ANSI_CHARSET;
	lf.lfOutPrecision=OUT_DEFAULT_PRECIS;
	lf.lfClipPrecision=CLIP_DEFAULT_PRECIS;
	lf.lfQuality=DEFAULT_QUALITY;
	lf.lfWeight=FW_NORMAL;
	lf.lfItalic=0;
	lf.lfPitchAndFamily=4;
	wcscpy (lf.lfFaceName, wface);

	FontColour=RGB(0,0,0);
	w = h = 0;
	ShowSIP = 1;
}

void SetFontProps ()
{
	lf.lfQuality=UseCleartype?CLEARTYPE_QUALITY:DEFAULT_QUALITY;
	lf.lfWeight=UseFontBold?FW_BOLD:FW_NORMAL;
	lf.lfItalic=UseFontItalic;
}

void MyApp::ReadIni()
{
	// read information from ini file
	SetDefs();
	ReadRegString("General","LastFile",(String&)LastFile);
	ReadRegInt("General","FiltIndex",FiltIndex);
	ReadRegString("General","LastGameFile",(String&)LastGameFile);
	ReadRegInt("General","GameFiltIndex",GameFiltIndex);

	ReadRegLong ("Font","Size",lf.lfHeight);
	String Face (LF_FACESIZE);
	ReadRegString("Font","Name",Face);
	if (Face.Len ())
	{
		WS wface = Face;
		wcscpy (lf.lfFaceName, wface);
	}
	ReadRegLong("Font","Colour",(long&) FontColour);
	ReadRegInt("Font", "Cleartype", UseCleartype);
	ReadRegInt("Font", "Bold", UseFontBold);
	ReadRegInt("Font", "Italic", UseFontItalic);
	ReadRegInt("General", "ShowSIP", ShowSIP);
	ReadRegInt("Graphics", "ShowPicture", ShowPicture);
	ReadRegInt("Graphics", "ScalePicture", ScalePicture);
	ReadRegInt("Graphics", "ColourScheme", ColourScheme);
	ReadRegInt("General", "ExpandAbbreviations", ExpandAbbreviations);
	ReadRegInt("General", "WriteLog", WriteLog);
	ReadRegString("General", "LogDir", LogDir);
	ReadRegInt("General", "ShowBlinker", ShowBlinker);

	SetFontProps ();
}

void MyApp::WriteIni()
{
 	char *cface = StringFromWS (lf.lfFaceName);
 	String face = cface;
 	delete [] cface;
	// write information to ini file
	WriteRegString("General","LastFile",LastFile);
	WriteRegInt("General","FiltIndex",FiltIndex);
	WriteRegString("General","LastGameFile",LastGameFile);
	WriteRegInt("General","GameFiltIndex",GameFiltIndex);
	WriteRegLong("Font","Size",(long) lf.lfHeight);
 	WriteRegString("Font", "Name", face);
	WriteRegLong("Font","Colour",(long) FontColour);
	WriteRegInt("Font", "Cleartype", UseCleartype);
	WriteRegInt("Font", "Bold", UseFontBold);
	WriteRegInt("Font", "Italic", UseFontItalic);
	WriteRegInt("General", "ShowSIP", ShowSIP);
	WriteRegInt("Graphics", "ShowPicture", ShowPicture);
	WriteRegInt("Graphics", "ScalePicture", ScalePicture);
	WriteRegInt("Graphics", "ColourScheme", ColourScheme);
	WriteRegInt("General", "ExpandAbbreviations", ExpandAbbreviations);
	WriteRegInt("General", "WriteLog", WriteLog);
	WriteRegString("General", "LogDir", LogDir);
	WriteRegInt("General", "ShowBlinker", ShowBlinker);
}

MyApp::MyApp(char *Name) : App(Name,Ini)
{
	ReadIni();
};

void MyApp::InitMainWindow()
{
	MainWindow=new ::MainWindow(0,MainWinTitle);

	if (has_aygshell)
	{
		int iDelta;
		SIPINFO si = {0};

		si.cbSize = sizeof(si);
		(*impSHSipInfo)(SPI_GETSIPINFO, 0, &si, 0);

		//Consider the menu at the bottom, please.
		iDelta = (si.fdwFlags & SIPF_ON) ? 0 : MENUBAR_HEIGHT;

		((Window *) MainWindow)->x = CW_USEDEFAULT;
		((Window *) MainWindow)->y = CW_USEDEFAULT;
		((Window *) MainWindow)->w = CW_USEDEFAULT; //si.rcVisibleDesktop.right  - si.rcVisibleDesktop.left;
		((Window *) MainWindow)->h = CW_USEDEFAULT; //si.rcVisibleDesktop.bottom - si.rcVisibleDesktop.top - iDelta;
	}
}

MyApp::~MyApp()
{
	WriteIni();
}

void ShowSip (BOOL Show)
{
	if (has_aygshell)
	{
		SIPINFO si;
		memset (&si, 0, sizeof (si));
		si.cbSize = sizeof (si);

		(*impSHSipInfo) (SPI_GETSIPINFO, 0, &si, 0);
		si.fdwFlags = (si.fdwFlags & (Show ? ~SIPF_OFF : ~SIPF_ON)) |
						(Show ? SIPF_ON : SIPF_OFF);
		(*impSHSipInfo) (SPI_SETSIPINFO, 0, &si, 0);
	}
	else
	{
		SipShowIM (Show ? SIPF_ON : SIPF_OFF);
	}
}

void MyApp::FirstIn()
{
	((::MainWindow*) MainWindow)->CmOpen();
}

void InitAygShell ()
{
	HINSTANCE hDll = LoadLibrary (_T("aygshell.dll"));

	if (hDll != NULL)
	{
		impSHSipInfo		= (tShSipInfo *) GetProcAddress (hDll, L"SHSipInfo");
		impSHCreateMenuBar	= (tShCreateMenuBar *) GetProcAddress (hDll, L"SHCreateMenuBar");

		if ((impSHSipInfo != NULL) &&
			(impSHCreateMenuBar != NULL))
		{
			has_aygshell = TRUE;
		}
		else
		{
			FreeLibrary (hDll);
		}
	}
}

// main function, called from WinMain()
int Main()
{
	HWND Window;
	WS wWinTitle = MainWinTitle;
	WS wAppName = AppName;

	if (Window = FindWindow (wAppName, wWinTitle))
	{
		SetForegroundWindow (Window);
		return 0;
	}

	InitAygShell ();

	// create and run application
	return MyApp(AppName).Run();
}
