#include <stringc.h>
#include <screenstring.h>
#include <inputoutput.h>
#include <screen.h>
#include <level9.h>
#include <ctype.h>

String Output;

Array<ScreenString> screenlines;
int topLine = 0;
int cont_lines = 0;
int after_input = 1;
int Caret;

int ExpandAbbreviations = 1;

typedef struct
{
	char abbrev;
	char *expand;
} Abbreviations;

Abbreviations abbreviations [] = {
	{'c', "close"}, {'k', "attack"}, {'q', "quit"}, {'x', "examine"},
	{'g', "again"}, {'l', "look"}, {'r', "drop"}, {'y', "yes"},
	{'i', "inventory"}, {'p', "open"}, {'t', "take"}, {'z', "wait"},
	{0, 0}};

SimpleList<int> InputChars;
String Hash(20);

SimpleList<String> History;
int HistoryLines=0;
#define NUMHIST 10

int iPos;

void DrawScreenString (ScreenString &sstr)
{
	for (int i = 0; i < sstr.numLines; i++)
	{
		ScreenString ss=sstr.GetScreenStringLine (i);
		DrawOneLine (ss.x, ss.y, ss.dx, ss.dy, ss.GetString ());
	}
}

void CleanScreenString (ScreenString &sstr)
{
  UnDrawRect (sstr.x, sstr.y, sstr.dx, sstr.dy);
}

void DeleteUnvisibleStrings (void)
{
	for (int i = 0; i < screenlines.GetSize (); i++)
		//if (!IsScreenStringVisible (*(screenlines[i])))
		if (!(screenlines[i] -> Visible ()))
		{
			screenlines.Delete (i);
			DeleteUnvisibleStrings ();
			return;
		}
}

void Scroll ()
{
	int i;

	PreScrollWindow ();

	for (i = 0; i < screenlines.GetSize (); i++)
		screenlines[i]->y--;

	DeleteUnvisibleStrings ();

	PostScrollWindow ();
}

void DeleteLastScreenString ()
{
	int last = screenlines.GetSize () - 1;

	ScreenString *slast = screenlines [last];

	CleanScreenString (*slast);

	topLine -= slast->numLines;
	if (topLine < 0)
		topLine = 0;

	screenlines.Delete (last);
}

ScreenString *NewScreenString (String &str)
{
	return new ScreenString (str, 0, topLine);
}

void AddScreenString (ScreenString *sstr)
{
	screenlines.Add (sstr);

	topLine += sstr->numLines;

	while (!sstr->FullyVisible ())
	{
		topLine--;
		Scroll ();
	}

	DrawScreenString (*sstr);

	PostAddScreenString ();
}

void AddScreenString (String &str)
{
	AddScreenString (NewScreenString (str));
}

void ReplaceLastScreenString (String &str)
{
	DeleteLastScreenString ();
	AddScreenString (str);
}

void RecreateScreenLines ()
{
	Array<ScreenString> array = screenlines;
	int i;

	for (i = 0; i < screenlines.GetSize (); i++)
		CleanScreenString (*screenlines[i]);

	screenlines.Clear ();

	topLine = 0;

	for (i = 0; i < array.GetSize (); i++)
		AddScreenString (array [i]->GetString ());
}

String GetWordAtPos (int x, int y)
{
	String word;

	int i = screenlines.GetSize ();

	while (!word.Len () && i--)
	{
		word = screenlines [i]->GetWord (x, y);
	}

	return word;
}
void DrawStrings (void)
{
	for (int i = 0; i < screenlines.GetSize (); i++)
		DrawScreenString (*(screenlines[i]));
}

String GetLastScreenString ()
{
	return screenlines[screenlines.GetSize () - 1]->GetString ();
}

void SetLastCaret ()
{
	if (Caret)
	{
		int last = screenlines.GetSize () - 1;

		SetCaret (screenlines [last]->GetCaretX (iPos),
				  screenlines [last]->GetCaretY (iPos));
	}
}

void AddWordAtPos (int x, int y)
{
	String word;
	char Char;

	word = GetWordAtPos (x, y);

	if (!word.Empty ())
	{
		InputChars.AddTail (' ');
		for (int j = 0; j < word.Len (); j++)
		{
			Char = word [j];

			if (!((Char >= 'A') && (Char <= 'Z')) &&
				!((Char >= 'a') && (Char <= 'z')) &&
				!((Char >= '0') && (Char <= '9')))
				break;
			else
				InputChars.AddTail (word [j]);
		}
		InputChars.AddTail (' ');
	}
}

void CancelInput()
{
	Hash = "";

	if (Caret) InputChars.AddTail (KEY_CANCEL_INPUT);
}

void ReplaceInput (const String &h)
{
	if (Caret)
	{
		Hash=h;
		InputChars.AddTail(KEY_REPLACE_INPUT);
	}
}

void print_string (String &str)
{
	for (int i = 0; i < str.Length; i++)
		os_printchar (str [i]);

	os_printchar ('\r');
}

int os_input(char*ibuff,int size)
{
	slIterator<String> HistPtr(History);
	if (HistPtr()) HistPtr--; // force out

	String Prompt = GetLastScreenString ();
	Prompt << String (" ");

	String Input = Prompt;
	String temp;

	int last;
	int start = Input.Len ();
	int InputChar;
	int i;
	
	cont_lines = 0;
	after_input = 1;

	iPos = start;
	Caret = 1;
	MakeCaret ();

	while (1)
	{
		last = screenlines.GetSize () - 1;
		HideCaret ();

		ReplaceLastScreenString (String () << Input << '.');

		SetLastCaret ();

		ShowCaret ();
		
		while (!(InputChar = os_getchar (10)) || (InputChar == -3));

		if (InputChar == KEY_REPLACE_INPUT) // Hash
		{
			ReplaceLastScreenString (String () << Prompt << Hash << '.');

			Input = Prompt;
			Input << Hash;
			InputChar = '\r';
		}

		if (InputChar == KEY_TIMEOUT)
			continue;
		else if (InputChar == KEY_CANCEL_INPUT) // CancelInput
		{
			KillCaret();
			Caret=0;

			return 0;
		}
		else if ((InputChar=='\r') || (InputChar == '\n'))
		{
			if (!strcmp (Input + start, "#abbreviations"))
			{
				KillCaret();
				Caret=0;

				temp = "Abbreviations are:\r";

				i = 0;
				while (abbreviations [i].abbrev)
				{
					if (i)
						temp << ", ";
					temp << abbreviations [i].abbrev << "->"
						 << abbreviations [i].expand;
					i++;
				}

				print_string (temp);

				os_flush ();

				print_string (String () << Prompt << '.');
				
				Input = Prompt;
				iPos = start;
				cont_lines = 0;
				after_input = 1;
				Caret = 1;
				MakeCaret ();
				SetLastCaret ();
				continue;
			}

			int len = strlen (Input + start);
			if (len > 0)
			{
				temp = "";

				int pos = start;

				KillCaret();
				Caret=0;

				if (ExpandAbbreviations && Input [pos] &&
					((Input [pos + 1] == '\0') || (Input [pos + 1] == ' ')))
				{
					i = 0;
					while (abbreviations [i].abbrev)
					{
						if (Input [start] == abbreviations [i].abbrev)
						{
							temp = abbreviations [i].expand;
							pos++;
							break;
						}
						i++;
					}
				}

				temp << Input + pos;
				os_write_log_string (String () << "Input> " << temp);

				strcpy(ibuff, temp); // >500 chrs??

				ReplaceLastScreenString (String () << Prompt << temp << '.');

				HistPtr.Last();
				if (*ibuff && (!HistPtr() || HistPtr.Get()!=ibuff))
				{
					History.AddTail(String(ibuff));
					if (HistoryLines==NUMHIST)
						delete History.GetFirst();
					else
						HistoryLines++;
				}
				return 1;
			}
			else
				continue;
		}
		else if (InputChar == KEY_DELETE_CODE)
		{
			if (iPos < Input.Len ())
			Input.Remove (iPos, 1);
		}
		else if (InputChar == KEY_BACKSPACE_CODE)
		{
				if (iPos > start)
			Input.Remove (--iPos, 1);
		}
		else if (InputChar == KEY_UP_CODE)
		{
			if (!HistPtr())
				HistPtr.Last();
			else
			{
				--HistPtr;
				if (!HistPtr()) HistPtr.First();
			}
			if (HistPtr())
			{
				Input = Prompt;
				Input << HistPtr.Get();
				iPos=Input.Len ();
			}
		}
		else if (InputChar == KEY_DOWN_CODE)
		{
			if (HistPtr())
			{
				++HistPtr;
				if (!HistPtr())
					HistPtr.Last();
				else
				{
					Input = Prompt;
					Input < HistPtr.Get();
					iPos = Input.Len ();
				}
			}
		}
		else if (InputChar == KEY_LEFT_CODE)
		{
			if (iPos > start) iPos--;
		}
		else if (InputChar == KEY_RIGHT_CODE)
		{
			if (iPos < Input.Len()) iPos++;
		}
		else if (InputChar == KEY_END_CODE)
		{
			iPos = Input.Len ();
		}
		else if (InputChar == KEY_HOME_CODE)
		{
			iPos = start;
		}
		else if (InputChar == KEY_ESCAPE_CODE)
		{
			Input [start] = 0;
			iPos = start;
		}
		else
		{
			if (((InputChar > ' ') && (InputChar < 256)) ||
				((InputChar == ' ') && (iPos > start) && (Input [iPos - 1] != ' ')))
			Input.Insert (InputChar, iPos++);
		}
	}

	return 0;
}

void os_flush (void)
{
	if (!Output.Empty ())
	{
		ScreenString *sstr = NewScreenString (Output);
		int lines = sstr->numLines;
		delete sstr;

		if (cont_lines >= ScreenString::MaxLines - lines)
		{
			String _ = String ("[Press a key]");
			AddScreenString (_);
			while (!os_getchar (60*1000)); // 1 minute timeout
			ReplaceLastScreenString (Output);

			cont_lines = ScreenString::MaxLines - screenlines [screenlines.GetSize () - 1] -> y;

			if (cont_lines < 0)
				cont_lines = 0;
		}
		else
		{
			cont_lines += lines;
			AddScreenString (Output);
		}

		os_write_log_string (Output);

		Output = "";
	}
}

void os_printchar(char c)
{
	if (!after_input || (c != '\r'))
		Output << c;
	after_input = 0;

	if (c == ' ')
	{
		ScreenString *sstr = NewScreenString (Output);

		while (sstr->numLines > 1)
		{
			Output = sstr->GetStringFirstLine ();
			os_flush ();
			Output = sstr->GetStringFromLine (1);
			sstr->SetString (Output);
			Output << c;
		}
		delete sstr;
	}
	else if (c == '\r')
		os_flush ();
}
