/***********************************************************************\
*
* Level 9 interpreter
* Version 4.0
* Copyright (c) 1996 Glen Summers
* Copyright (c) 2002,2003 Glen Summers and David Kinder
* Copyright (c) 2005 Glen Summers, David Kinder, Alan Staniforth,
* Simon Baldwin and Dieter Baron
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
*
\***********************************************************************/

#ifndef LEVEL9_H
#define LEVEL9_H

#include <stdio.h>
#include <winbase.h>

typedef unsigned char L9BYTE;
typedef unsigned short L9UINT16;
typedef unsigned long L9UINT32;
typedef int L9BOOL;

#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif

#define LISTAREASIZE 0x800
#define STACKSIZE 1024
#define V1FILESIZE 0x600

#ifndef MAX_PATH
#define MAX_PATH 256
#endif

typedef struct
{
	L9UINT32 Id;
	L9UINT16 codeptr,stackptr,listsize,stacksize,filenamesize,checksum;
	L9UINT16 vartable[256];
	L9BYTE listarea[LISTAREASIZE];
	L9UINT16 stack[STACKSIZE];      
	char filename[MAX_PATH];
} GameState;

typedef enum
{
	NO_BITMAPS,
	AMIGA_BITMAPS,
	PC1_BITMAPS,
	PC2_BITMAPS,
	C64_BITMAPS,
	BBC_BITMAPS,
	CPC_BITMAPS,
	MAC_BITMAPS,
	ST1_BITMAPS,
	ST2_BITMAPS,
} BitmapType;

typedef struct
{
	L9BYTE red, green, blue;
} Colour;

typedef struct
{
  L9UINT16 width, height;
  L9BYTE* bitmap;
  Colour palette[32];
  L9UINT16 npalette;
} Bitmap;

#define MAX_BITMAP_WIDTH 512
#define MAX_BITMAP_HEIGHT 216

#if (defined(_Windows) || defined(__MSDOS__) || defined (_WIN32) || defined (__WIN32__)) && (defined (_X86_) || defined (x86))
	#define L9WORD(x) (*(L9UINT16*)(x))
	#define L9SETWORD(x,val) (*(L9UINT16*)(x)=(L9UINT16)val)
	#define L9SETDWORD(x,val) (*(L9UINT32*)(x)=val)
#else
	#define L9WORD(x) (*(x) + ((*(x+1))<<8))
	#define L9SETWORD(x,val) *(x)=(L9BYTE) val; *(x+1)=(L9BYTE)(val>>8);
	#define L9SETDWORD(x,val) *(x)=(L9BYTE)val; *(x+1)=(L9BYTE)(val>>8); *(x+2)=(L9BYTE)(val>>16); *(x+3)=(L9BYTE)(val>>24);
#endif

#ifdef __cplusplus
extern "C" {
#endif

/* routines provided by os dependent code */
void os_printchar(char c);
L9BOOL os_input(char* ibuff, int size);
char os_readchar(int millis);
L9BOOL os_stoplist(void);
void os_flush(void);
L9BOOL os_save_file(L9BYTE* Ptr, int Bytes);
L9BOOL os_load_file(L9BYTE* Ptr, int* Bytes, int Max);
L9BOOL os_get_game_file(char* NewName, int Size);
void os_set_filenumber(char* NewName, int Size, int n);
void os_graphics(int mode);
void os_cleargraphics(void);
void os_setcolour(int colour, int index);
void os_drawline(int x1, int y1, int x2, int y2, int colour1, int colour2);
void os_fill(int x, int y, int colour1, int colour2);
void os_show_bitmap(int pic, int x, int y);

/* routines provided by level9 interpreter */
L9BOOL LoadGame(char* filename, char* picname);
L9BOOL RunGame(void);
void StopGame(void);
void RestoreGame(char* filename);
void FreeMemory(void);
void GetPictureSize(int* width, int* height);
L9BOOL RunGraphics(void);
void SetScaleGraphics(L9BOOL scale);

/* bitmap routines provided by level9 interpreter */
#ifdef BITMAP_DECODER
BitmapType DetectBitmaps(char* dir);
Bitmap* DecodeBitmap(char* dir, BitmapType type, int num, int x, int y);
#endif

#ifdef NEED_STRICMP_PROTOTYPE
int stricmp(const char* str1, const char* str2);
int strnicmp(const char* str1, const char* str2, size_t n);
#endif

/* stubs for function not found in WindowsCE */
#define time level9_time_stub
#define calloc level9_calloc
#ifndef isalnum
#define isalnum level9_isalnum
#endif
#define stricmp level9_stricmp
#define strnicmp level9_strnicmp

__inline int level9_time_stub (void *pvoid) { return Random (); }
__inline void* level9_calloc (size_t num, size_t size)
{
	void *result = malloc (num * size);
	memset (result, 0, num * size);

	return result;
}

__inline int level9_isalnum (int c)
{
	return ((c >= '0' && c <= '9') || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'));
}

__inline int level9_chrcasechr (int ch1, int ch2)
{
	int ch1_up, ch2_up;

	if (ch1 >= 'a' && ch1 <= 'z') ch1_up = ch1 + 'A' - 'a';
	if (ch2 >= 'a' && ch2 <= 'z') ch2_up = ch2 + 'A' - 'a';

	if (ch1_up == ch2_up)
		return 0;
	else
		if (ch1_up < ch2_up)
			return -1;
		else
			return 1;
}

__inline int level9_strnicmp (const char *str1, const char *str2, size_t n)
{
	int cmp;
	size_t min = n;
	if (n < strlen (str1)) min = strlen (str1);
	if (n < strlen (str2)) min = strlen (str2);

	while (min--)
		if ((cmp = level9_chrcasechr (str1 [min], str2 [min])) != 0)
			return cmp;

	return 0;
}

__inline int level9_stricmp (const char *str1, const char *str2)
{
	return level9_strnicmp (str1, str2, strlen (str1));
}

#ifdef __cplusplus
}
#endif

#endif
